import pytest
import warnings
import re
import io
import logging
import warnings
from types import ModuleType
import numpy as np
from mpi4py import MPI
from hymd.input_parser import (
    Config,
    _find_unique_names,
    _setup_type_to_name_map,
    read_config_toml,
    parse_config_toml,
    check_n_particles,
    check_max_molecule_size,
    check_bonds,
    check_angles,
    check_chi,
    check_box_size,
    check_integrator,
    check_thermostat_coupling_groups,
    check_cancel_com_momentum,
    check_start_and_target_temperature,
    check_n_print,
    check_tau,
    check_mass,
    check_domain_decomposition,
    check_name,
    check_config,
    check_hamiltonian,
    check_charges,
    check_n_flush,
)


def test_input_parser_read_config_toml(config_toml):
    config_toml_file, config_toml_str = config_toml
    file_content = read_config_toml(config_toml_file)
    assert config_toml_str == file_content


def test_input_parser_file(config_toml):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    assert isinstance(config, Config)
    assert isinstance(str(config), str)


@pytest.mark.mpi()
def test_input_parser_file_check_n_particles(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    n_particles_config = config.n_particles
    indices = np.empty((n_particles_config + 1,))
    indices_slices = np.array_split(indices, MPI.COMM_WORLD.Get_size())
    indices_take = indices_slices[MPI.COMM_WORLD.Get_rank()]

    if MPI.COMM_WORLD.Get_rank() == 0:
        with pytest.warns(Warning) as recorded_warning:
            config_ = check_n_particles(config, indices_take)
            assert config_.n_particles == 10001
            message = recorded_warning[0].message.args[0]
            log = caplog.text
            assert all([(s in message) for s in ("10000", "not", "10001")])
            assert all([(s in log) for s in ("10000", "not", "10001")])
    else:
        with warnings.catch_warnings() as recorded_warning:
            warnings.simplefilter("error")
            config_ = check_n_particles(config, indices_take)
            assert config_.n_particles == 10001

    caplog.clear()
    indices = np.empty((n_particles_config,))
    indices_slices = np.array_split(indices, MPI.COMM_WORLD.Get_size())
    indices_take = indices_slices[MPI.COMM_WORLD.Get_rank()]

    with warnings.catch_warnings() as recorded_warning:
        warnings.simplefilter("error")
        check_n_particles(config, indices_take)
        assert not recorded_warning
        assert not caplog.text
    MPI.COMM_WORLD.Barrier()


@pytest.mark.mpi()
def test_input_parser_check_optionals(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config_toml_no_n_particles = re.sub(
        "n_particles[ \t]*=[ \t]*[0-9]+", "", config_toml_str
    )
    config = parse_config_toml(config_toml_no_n_particles)
    assert config.n_particles is None
    indices = np.empty((100,))
    indices_slices = np.array_split(indices, MPI.COMM_WORLD.Get_size())
    indices_take = indices_slices[MPI.COMM_WORLD.Get_rank()]

    config = check_n_particles(config, indices_take)
    assert config.n_particles == 100
    if MPI.COMM_WORLD.Get_rank() == 0:
        assert all([(s in caplog.text) for s in ("No", "n_particles", "100")])
    caplog.clear()

    config_toml_no_max_molecule_size = re.sub(
        "max_molecule_size[ \t]*=[ \t]*[0-9]+", "", config_toml_str
    )
    config = parse_config_toml(config_toml_no_max_molecule_size)
    assert config.max_molecule_size is None
    config = check_max_molecule_size(config)
    if MPI.COMM_WORLD.Get_rank() == 0:
        assert all([(s in caplog.text) for s in ("No", "max_molecule_size", "201")])  # noqa: E501
    caplog.clear()

    config_toml_wrong_max_molecule_size = re.sub(
        "max_molecule_size[ \t]*=[ \t]*[0-9]+", "max_molecule_size = 0",
        config_toml_str
    )

    if MPI.COMM_WORLD.Get_rank() == 0:
        with pytest.warns(Warning) as recorded_warning:
            config = parse_config_toml(config_toml_wrong_max_molecule_size)
            config = check_max_molecule_size(config)
            assert config.max_molecule_size == 201
            message = recorded_warning[0].message.args[0]
            log = caplog.text
            assert all([(s in message) for s in ("must be", "integer", "201")])
            assert all([(s in log) for s in ("must be", "integer", "201")])
    else:
        with warnings.catch_warnings() as recorded_warning:
            warnings.simplefilter("error")
            config = parse_config_toml(config_toml_wrong_max_molecule_size)
            config = check_max_molecule_size(config)
            assert config.max_molecule_size == 201

    MPI.COMM_WORLD.Barrier()


def _add_to_config(config_str, new_str, header_str):
    sio = io.StringIO(config_str)
    sio_new = []
    header_flag = False
    for line in sio:
        if line.strip().startswith(f"{header_str} ="):
            header_flag = True
        if header_flag and line.strip() == "]":
            header_flag = False
            sio_new.append(new_str)
        sio_new.append(line.rstrip())
    return "\n".join(s for s in sio_new)


def _remove_from_config(config_str, remove_str):
    sio = io.StringIO(config_str)
    sio_new = []
    for line in sio:
        if remove_str not in line.strip():
            sio_new.append(line.rstrip())
    return "\n".join(s for s in sio_new)


def _change_in_config(config_str, old_line, new_line):
    sio = io.StringIO(config_str)
    sio_new = []
    for line in sio:
        if line.strip().startswith(old_line.strip()):
            sio_new.append(new_line.rstrip())
        else:
            sio_new.append(line.rstrip())
    return "\n".join(s for s in sio_new)


@pytest.mark.mpi()
def test_input_parser_check_bonds(config_toml, dppc_single, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    _, _, names, _, _, _ = dppc_single
    solvent_names = np.empty((8), dtype="a5")
    solvent_names.fill("W")
    names = np.concatenate((names, solvent_names))

    name_slices = np.array_split(names, MPI.COMM_WORLD.Get_size())
    names_take = name_slices[MPI.COMM_WORLD.Get_rank()]

    add_bonds = [
        '  ["A", "C", 0.47, 1250.0],',
        '  ["A", "A", 0.47, 1250.0],',
        '  ["P", "B", 0.47, 1250.0],',
        '  ["A", "A", 0.47, 1250.0],',
        '  ["A", "B", 0.47, 1250.0],',
    ]
    warn_strs = [
        ["A--C", "no A"],
        ["A--A", "no A"],
        ["P--B", "no B"],
        ["A--A", "no A"],
        ["A--B", "neither A, B"],
    ]

    for a, w in zip(add_bonds, warn_strs):
        added_bonds_toml_str = _add_to_config(config_toml_str, a, "bonds")
        config = parse_config_toml(added_bonds_toml_str)

        if MPI.COMM_WORLD.Get_rank() == 0:
            with pytest.warns(Warning) as recorded_warning:
                config = check_bonds(config, names_take)
                message = recorded_warning[0].message.args[0]
                log = caplog.text
                assert all([(s in message) for s in w])
                assert all([(s in log) for s in w])
        else:
            with warnings.catch_warnings() as recorded_warning:
                warnings.simplefilter("error")
                config = check_bonds(config, names_take)

        caplog.clear()
    MPI.COMM_WORLD.Barrier()


@pytest.mark.mpi()
def test_input_parser_check_angles(config_toml, dppc_single, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    _, _, names, _, _, _ = dppc_single
    solvent_names = np.empty((8), dtype="a5")
    solvent_names.fill("W")
    names = np.concatenate((names, solvent_names))

    name_slices = np.array_split(names, MPI.COMM_WORLD.Get_size())
    names_take = name_slices[MPI.COMM_WORLD.Get_rank()]

    add_angles = [
        '  ["A", "C", "C", 115.9, 25.0],',
        '  ["A", "A", "A", 158.5, 25.0],',
        '  ["P", "B", "C",   1.5, 25.0],',
        '  ["A", "A", "B",  99.8, 25.0],',
        '  ["A", "B", "K", 119.8, 25.0],',
    ]
    warn_strs = [
        ["A--C--C", "no A"],
        ["A--A--A", "no A"],
        ["P--B--C", "no B"],
        ["A--A--B", "no A, B"],
        ["A--B--K", "no A, B, K"],
    ]

    for a, w in zip(add_angles, warn_strs):
        added_angles_toml_str = _add_to_config(
            config_toml_str, a, "angle_bonds"
        )
        config = parse_config_toml(added_angles_toml_str)

        if MPI.COMM_WORLD.Get_rank() == 0:
            with pytest.warns(Warning) as recorded_warning:
                config = check_angles(config, names_take)
                message = recorded_warning[0].message.args[0]
                log = caplog.text
                assert all([(s in message) for s in w])
                assert all([(s in log) for s in w])
        else:
            with warnings.catch_warnings() as recorded_warning:
                warnings.simplefilter("error")
                config = check_angles(config, names_take)

        caplog.clear()
    MPI.COMM_WORLD.Barrier()


@pytest.mark.mpi()
def test_input_parser_check_chi(config_toml, dppc_single, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    for c in config.chi:
        if c.atom_1 == "G" and c.atom_2 == "W":
            assert c.interaction_energy == pytest.approx(4.53, abs=1e-2)
        elif c.atom_1 == "P" and c.atom_2 == "C":
            assert c.interaction_energy == pytest.approx(14.72, abs=1e-2)

    _, _, names, _, _, _ = dppc_single
    solvent_names = np.empty((8), dtype="a5")
    solvent_names.fill("W")
    names = np.concatenate((names, solvent_names))

    name_slices = np.array_split(names, MPI.COMM_WORLD.Get_size())
    names_take = name_slices[MPI.COMM_WORLD.Get_rank()]

    add_chi = [
        '  ["A", "C", 1.2398],',
        '  ["A", "A", -9.0582],',
        '  ["P", "B", -8.8481],',
        '  ["A", "A", 3.1002],',
        '  ["A", "B", 2.7815],',
    ]
    warn_strs = [
        ["A--C", "no A"],
        ["A--A", "no A"],
        ["B--P", "no B"],
        ["A--A", "no A"],
        ["A--B", "neither A, B"],
    ]

    for a, w in zip(add_chi, warn_strs):
        added_chi_toml_str = _add_to_config(config_toml_str, a, "chi")
        config = parse_config_toml(added_chi_toml_str)

        if MPI.COMM_WORLD.Get_rank() == 0:
            with pytest.warns(Warning) as recorded_warning:
                config = check_chi(config, names_take)
                message = recorded_warning[0].message.args[0]
                log = caplog.text
                assert all([(s in message) for s in w])
                assert all([(s in log) for s in w])
        else:
            with warnings.catch_warnings() as recorded_warning:
                warnings.simplefilter("error")
                config = check_chi(config, names_take)

        caplog.clear()

    remove_chi = [
        '["C", "W", 42.24],',
        '["N", "P", -9.34],',
        '["N", "C", 13.56],',
        '["P", "C", 14.72],',
    ]
    warn_strs = [
        ["C and W", "no chi interaction C--W", "Defaulting"],
        ["N and P", "no chi interaction N--P", "Defaulting"],
        ["C and N", "no chi interaction C--N", "Defaulting"],
        ["C and P", "no chi interaction C--P", "Defaulting"],
    ]

    for r, w in zip(remove_chi, warn_strs):
        removed_chi_toml_str = _remove_from_config(config_toml_str, r)
        config = parse_config_toml(removed_chi_toml_str)

        if MPI.COMM_WORLD.Get_rank() == 0:
            with pytest.warns(Warning) as recorded_warning:
                config = check_chi(config, names_take)
                message = recorded_warning[0].message.args[0]
                log = caplog.text
                assert all([(s in message) for s in w])
                assert all([(s in log) for s in w])
        else:
            with warnings.catch_warnings() as recorded_warning:
                warnings.simplefilter("error")
                config = check_chi(config, names_take)

        caplog.clear()
    MPI.COMM_WORLD.Barrier()


def test_input_parser_check_box_size(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    assert np.allclose(np.array([2.1598, 11.2498, 5.1009]), config.box_size, atol=1e-4)  # noqa: E501

    changed_box_toml_str = _change_in_config(
        config_toml_str, "box_size = [", "box_size = [2.25, -3.91, 4.11]"
    )
    config = parse_config_toml(changed_box_toml_str)
    input_box = np.array([2.25, -3.91, 4.11])

    with pytest.raises(ValueError) as recorded_error:
        _ = check_box_size(config, input_box)
        log = caplog.text
        assert all([(s in log) for s in ("Invalid", "box")])
    message = str(recorded_error.value)
    assert all([(s in message) for s in ("Invalid", "box")])

    input_box = np.ones(3)
    with pytest.raises(ValueError) as recorded_error:
        _ = check_box_size(config, input_box)
        log = caplog.text
        assert all([(s in log) for s in ("Box", "not", "match")])
    message = str(recorded_error.value)
    assert all([(s in message) for s in ("Box", "not", "match")])

    caplog.clear()


def test_input_parser_check_integrator(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    assert config.integrator == "respa"
    assert config.respa_inner == 5

    with pytest.warns(Warning) as recorded_warning:
        config.respa_inner = 1.0
        config = check_integrator(config)
        assert config.respa_inner == 1
        log = caplog.text
        message = recorded_warning[0].message.args[0]
        cmp_strings = ("Number of inner rRESPA time steps",
                       "specified as float")
        assert all([(s in message) for s in cmp_strings])
        assert all([(s in log) for s in cmp_strings])

    with pytest.raises(TypeError) as recorded_error:
        config.respa_inner = "test"
        _ = check_integrator(config)
    message = str(recorded_error.value)
    cmp_strings = ("Invalid number of inner rRESPA time steps", 
                   "Must be positive integer")
    assert all([(s in message) for s in cmp_strings])

    with pytest.raises(ValueError) as recorded_error:
        config.respa_inner = -1
        _ = check_integrator(config)
    message = str(recorded_error.value)
    cmp_strings = ("Invalid number of inner rRESPA time steps", 
                   "Must be positive integer")
    assert all([(s in message) for s in cmp_strings])

    with pytest.warns(Warning) as recorded_warning:
        config.integrator = "velocity-verlet"
        config.respa_inner = 2
        config = check_integrator(config)
        assert config.respa_inner == 1
        log = caplog.text
        message = recorded_warning[0].message.args[0]
        cmp_strings = ("Integrator type Velocity-Verlet specified",
                       "and inner rRESPA time steps set to",
                       "Using respa_inner = 1")
        assert all([(s in message) for s in cmp_strings])
        assert all([(s in log) for s in cmp_strings])

    changed_integrator_toml_str = _change_in_config(
        config_toml_str, "integrator = ", 'integrator = "rsjgyu"'
    )
    config = parse_config_toml(changed_integrator_toml_str)

    with pytest.raises(ValueError) as recorded_error:
        _ = check_integrator(config)
        if MPI.COMM_WORLD.Get_rank() == 0:
            log = caplog.text
            assert all([(s in log) for s in ("Invalid", "integrator")])
    message = str(recorded_error.value)
    assert all([(s in message) for s in ("Invalid", "integrator")])
    caplog.clear()


def test_input_parser_thermostat_coupling_groups(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    nn = (("N", "P"), ("G", "C"), ("W"))
    for i in range(3):
        for n in nn[i]:
            assert n in config.thermostat_coupling_groups[i]

    thermostat_coupling_groups_toml_str = _remove_from_config(
        config_toml_str, '["N", "P"],'
    )
    thermostat_coupling_groups_toml_str = _remove_from_config(
        thermostat_coupling_groups_toml_str, '["G", "C"],'
    )
    thermostat_coupling_groups_toml_str = _remove_from_config(
        thermostat_coupling_groups_toml_str, '["W"],'
    )
    thermostat_coupling_groups_toml_str = _add_to_config(
        thermostat_coupling_groups_toml_str,
        '        ["N", "P", "G", "C"],',
        "thermostat_coupling_groups",
    )
    thermostat_coupling_groups_toml_str = _add_to_config(
        thermostat_coupling_groups_toml_str,
        '        ["W"],',
        "thermostat_coupling_groups",
    )
    config = parse_config_toml(thermostat_coupling_groups_toml_str)
    nn = (("N", "P", "G", "C"), ("W"))
    for i in range(2):
        for n in nn[i]:
            assert n in config.thermostat_coupling_groups[i]

    thermostat_coupling_groups_toml_str = _remove_from_config(
        thermostat_coupling_groups_toml_str, '["N", "P", "G", "C"],'
    )
    thermostat_coupling_groups_toml_str = _add_to_config(
        thermostat_coupling_groups_toml_str,
        '        ["N", "P", "G"],',
        "thermostat_coupling_groups",
    )

    config = parse_config_toml(thermostat_coupling_groups_toml_str)
    config.unique_names = sorted(["N", "P", "G", "C", "W"])
    with pytest.raises(ValueError) as recorded_error:
        _ = check_thermostat_coupling_groups(config)
        log = caplog.text
        assert all([(s in log) for s in ("species C", "not specified")])

    message = str(recorded_error.value)
    assert all([(s in message) for s in ("species C", "not specified")])
    caplog.clear()

    thermostat_coupling_groups_toml_str = _remove_from_config(
        thermostat_coupling_groups_toml_str, '["N", "P", "G"],'
    )
    thermostat_coupling_groups_toml_str = _remove_from_config(
        thermostat_coupling_groups_toml_str, '["W"],'
    )
    thermostat_coupling_groups_toml_str = _add_to_config(
        thermostat_coupling_groups_toml_str,
        '        ["N", "P", "G", "C"],',
        "thermostat_coupling_groups",
    )
    thermostat_coupling_groups_toml_str = _add_to_config(
        thermostat_coupling_groups_toml_str,
        '        ["W", "P"],',
        "thermostat_coupling_groups",
    )
    config = parse_config_toml(thermostat_coupling_groups_toml_str)
    config.unique_names = sorted(["N", "P", "G", "C", "W"])
    with pytest.raises(ValueError) as recorded_error:
        _ = check_thermostat_coupling_groups(config)
        log = caplog.text
        assert all([(s in log) for s in ("species P", "specified", "multiple")])  # noqa: E501

    message = str(recorded_error.value)
    print(message)
    assert all([(s in message) for s in ("species P", "specified", "multiple")])  # noqa: E501
    caplog.clear()


def test_input_parser_check_cancel_com_momentum(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    for t in (
        1.1,
        "hello",
        MPI.COMM_WORLD,
        config_toml,
        [1],
    ):
        config.cancel_com_momentum = t
        with pytest.raises(ValueError) as recorded_error:
            _ = check_cancel_com_momentum(config)
            log = caplog.text
            assert all([(s in log) for s in ("not interpret", "an integer")])

        message = str(recorded_error.value)
        assert all([(s in message) for s in ("not interpret", "an integer")])
        caplog.clear()

    for t in (-1, -100, 0, False):
        config.cancel_com_momentum = t
        config_ = check_cancel_com_momentum(config)
        assert config_.cancel_com_momentum is False
    caplog.clear()


def test_input_parser_check_start_and_target_temperature(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    for t in (
        "hello",
        MPI.COMM_WORLD,
        config_toml,
        [1],
    ):
        config.start_temperature = t
        with pytest.raises(TypeError) as recorded_error:
            _ = check_start_and_target_temperature(config)
            log = caplog.text
            assert all([(s in log) for s in ("not interpret", "a number")])
        message = str(recorded_error.value)
        assert all([(s in message) for s in ("not interpret", "a number")])
        caplog.clear()

        config.target_temperature = t
        with pytest.raises(TypeError) as recorded_error:
            _ = check_start_and_target_temperature(config)
            log = caplog.text
            assert all([(s in log) for s in ("not interpret", "a number")])
        message = str(recorded_error.value)
        assert all([(s in message) for s in ("not interpret", "a number")])
        caplog.clear()

    config = parse_config_toml(config_toml_str)

    config.start_temperature = None
    assert check_start_and_target_temperature(config).start_temperature is False  # noqa: E501
    config.target_temperature = None
    assert check_start_and_target_temperature(config).target_temperature is False  # noqa: E501

    with pytest.warns(Warning) as recorded_warning:
        config.start_temperature = -6.2985252885781357
        config = check_start_and_target_temperature(config)
        assert config.start_temperature is False
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        assert all([(s in message) for s in ("to negative", "defaulting to")])
        assert all([(s in log) for s in ("to negative", "defaulting to")])
    caplog.clear()

    with pytest.warns(Warning) as recorded_warning:
        config.target_temperature = -0.000025892857873
        config = check_start_and_target_temperature(config)
        assert config.target_temperature is False
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        assert all([(s in message) for s in ("to negative", "defaulting to")])
        assert all([(s in log) for s in ("to negative", "defaulting to")])
    caplog.clear()


def test_input_parser_check_n_print(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config.n_print = None
    config = check_n_print(config)
    assert config.n_print == False

    with pytest.warns(Warning) as recorded_warning:
        config.n_print = 1.27
        config = check_n_print(config)
        assert config.n_print == 1
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_string = "n_print is a float"
        assert cmp_string in message
        assert cmp_string in log

    with pytest.raises(RuntimeError) as recorded_error:
        config.n_print = "test"
        _ = check_n_print(config)        
        log = caplog.text
        cmp_string = "invalid value for n_print"
        assert cmp_string in log
    message = str(recorded_error.value)
    cmp_string = "invalid value for n_print"
    assert cmp_string in message

    caplog.clear()


def test_input_parser_check_tau(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    with pytest.warns(Warning) as recorded_warning:
        config.tau = None
        config = check_tau(config)
        assert config.tau == pytest.approx(0.7)
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_string = "target temp specified but no tau"
        assert cmp_string in message
        assert cmp_string in log

    caplog.clear()

   
def test_input_parser_check_mass(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config = check_mass(config)
    assert config.mass == pytest.approx(72.0)

    config.mass = None
    config = check_mass(config)
    log = caplog.text
    assert config.mass == pytest.approx(72.0)
    cmp_string = "no mass specified, defaulting to 72.0"
    assert cmp_string in log

    with pytest.raises(TypeError) as recorded_error:
        config.mass = "test"
        _ = check_mass(config)
    message = str(recorded_error.value)
    cmp_string = "specified mass is invalid type"
    assert cmp_string in message

    with pytest.raises(ValueError) as recorded_error:
        config.mass = -1.0
        _ = check_mass(config)
    message = str(recorded_error.value)
    cmp_string = "invalid mass specified"
    assert cmp_string in message

    caplog.clear()


def test_input_parser_check_domain_decomposition(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config = check_domain_decomposition(config)
    assert config.domain_decomposition == False

    with pytest.warns(Warning) as recorded_warning:
        config.domain_decomposition = -1
        config = check_domain_decomposition(config)
        assert config.domain_decomposition == False
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_string = "negative domain_decomposition specified, using False"
        assert cmp_string in message
        assert cmp_string in log

    with pytest.warns(Warning) as recorded_warning:
        config.domain_decomposition = -1.1
        config = check_domain_decomposition(config)
        assert config.domain_decomposition == False
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_string = "negative domain_decomposition specified, using False"
        assert cmp_string in message
        assert cmp_string in log
    
    with pytest.warns(Warning) as recorded_warning:
        config.domain_decomposition = 1.0
        config = check_domain_decomposition(config)
        assert config.domain_decomposition == 1
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_strings = ("domain_decomposition", "is not an integer, using")
        assert all([(s in message) for s in cmp_strings])
        assert all([(s in log) for s in cmp_strings])

    with pytest.raises(ValueError) as recorded_error:
        config.domain_decomposition = "test"
        _ = check_domain_decomposition(config)
    message = str(recorded_error.value)
    cmp_strings = ("invalid value for domain_decomposition", 
                   "use an integer")
    assert all([(s in message) for s in cmp_strings])

    caplog.clear()


def test_input_parser_check_name(config_toml):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config = check_name(config)
    assert "example config.toml" in config.name

    config.name = None
    config = check_name(config)
    assert "sim" in config.name


def test_input_parser_check_config(config_toml, dppc_single):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    config.n_particles = 13
    charges = np.zeros(config.n_particles, dtype=np.float32)
    input_box = np.array(config.box_size)
    config.n_b = 1
    config.barostat_type = "berendsen"

    indices, _, names, _, _, _ = dppc_single

    indices = np.append(indices, [12])
    names = np.append(names, [b"W"])
    names_to_types = {"N":0, "P": 1, "G": 2, "C": 3, "W": 4}
    types = np.array([names_to_types[n.decode('UTF-8')] for n in names],
                     dtype=int)

    config = check_config(config, indices, names, types, charges, input_box)
    assert isinstance(config, Config)


def test_input_parser_check_hamiltonian(config_toml, caplog):
    caplog.set_level(logging.INFO)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config = check_hamiltonian(config)
    assert config.hamiltonian == "DefaultWithChi"

    with pytest.warns(Warning) as recorded_warning:
        config.hamiltonian = None
        config = check_hamiltonian(config)
        assert config.hamiltonian == "DefaultWithChi"
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_strings = ("No hamiltonian form specified", 
                       "defaulting to DefaultWithChi")
        assert all([(s in message) for s in cmp_strings])
        assert all([(s in log) for s in cmp_strings])

    with pytest.warns(Warning) as recorded_warning:
        config.hamiltonian = None
        config.chi = []
        config = check_hamiltonian(config)
        assert config.hamiltonian == "DefaultNoChi"
        message = recorded_warning[0].message.args[0]
        log = caplog.text
        cmp_strings = ("No hamiltonian form and no chi", 
                       "defaulting to DefaultNoChi")
        assert all([(s in message) for s in cmp_strings])
        assert all([(s in log) for s in cmp_strings])

    with pytest.raises(NotImplementedError) as recorded_error:
        config.hamiltonian = "test"
        _ = check_hamiltonian(config)
    message = str(recorded_error.value)
    cmp_strings = ("The specified Hamiltonian", 
                   "was not recognized as a valid Hamiltonian")
    assert all([(s in message) for s in cmp_strings])

    caplog.clear()


def test_input_parser__find_unique_names(config_toml, dppc_single):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    _, _, names, _, _, _ = dppc_single

    config = _find_unique_names(config, names)
    assert hasattr(config, "unique_names")
    assert isinstance(config.unique_names, list)
    assert all([(n in config.unique_names) for n in ["N","P","G","C"]])
    assert hasattr(config, "n_types")
    assert config.n_types == 4


def test_input_parser__setup_type_to_name_map(config_toml, dppc_single):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    _, _, names, _, _, _ = dppc_single

    names_to_types = {"N":0, "P": 1, "G": 2, "C": 3}
    types = np.array([names_to_types[n.decode('UTF-8')] for n in names],
                     dtype=int)

    config = _setup_type_to_name_map(config, names, types)
    assert hasattr(config, "name_to_type_map")
    assert isinstance(config.name_to_type_map, dict)
    assert all([(k in config.name_to_type_map.keys()) 
                 for k in names_to_types.keys()])
    assert hasattr(config, "type_to_name_map")
    assert isinstance(config.type_to_name_map, dict)
    assert all([(v in config.type_to_name_map.keys()) 
                 for v in names_to_types.values()])


@pytest.mark.mpi()
def test_input_parser_check_charges(config_toml, caplog):
    caplog.set_level(logging.WARNING)
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    config.coulombtype = "PIC_Spectral"
    config.dielectric_const = 80.

    charges = np.array(
        [1.0, 0.0, 0.5, 0.2, -0.3, 0.0, 0.3, 0.0, -0.5, 0.0, -0.5, 0.99, -0.2,
        0.3, 0.5, 0.0, 0.0, -0.3, 0.0, 0.0, -0.99, -1.0],
        dtype=float
    )
    # split charges across ranks
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    n_charges = charges.shape[0] // size
    ind_rank = {}
    for i in range(size):
        ind_rank[i] = i * n_charges
    ind_rank[size] = size * n_charges + np.mod(charges.shape[0], size)

    rank_charges = charges[ind_rank[rank]:ind_rank[rank+1]]

    newconf = check_charges(config, rank_charges, comm=comm) # warnings are not expected
    assert len(caplog.text) == 0
    assert newconf.self_energy != 0. # TODO: compare to the right value

    # change charges to generate a warning
    charges[1] = 1000.
    rank_charges = charges[ind_rank[rank]:ind_rank[rank+1]]

    with pytest.warns(Warning) as recorded_warning:
        newconf = check_charges(config, rank_charges)
        # only rank 0 gives the warning
        if rank == 0:
            message = recorded_warning[0].message.args[0]
            log = caplog.text
            cmp_strings = ("Charges in the input file do not sum to zero.", 
                           "Total charge is ")
            assert all([(s in message) for s in cmp_strings])
            assert all([(s in log) for s in cmp_strings])
            total_charge = float(log.split()[-1][:-1])
            assert total_charge == pytest.approx(charges[1])
        # but others should give a warn so pytest.warns does not fail with:
        # Failed: DID NOT WARN. No warnings of type..
        else:
            warnings.warn("give a warning so other ranks do not fail")
            assert len(recorded_warning) == 1


def test_input_parser_check_n_flush(config_toml):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    config = check_n_flush(config)

    assert config.n_flush == (10000 // config.n_print)

    config.n_flush = 1234

    config = check_n_flush(config)

    assert config.n_flush == 1234



import pytest
import logging
import numpy as np
import os
import argparse
from mpi4py import MPI
from argparse import Namespace
from hymd.configure_runtime import configure_runtime, extant_file
from hymd.input_parser import Config


def test_extant_file(tmp_path, caplog):
    caplog.set_level(logging.INFO)
    with pytest.raises(argparse.ArgumentTypeError) as recorded_error:
        _ = extant_file("inexistant_test_file.txt")
        log = caplog.text
        assert "does not exist" in log
    message = str(recorded_error.value)
    assert "does not exist" in message
    
    file_path = os.path.join(tmp_path, "test.txt")
    with open(file_path, 'w') as f:
        f.write("test")

    ret_path = extant_file(file_path)
    assert ret_path == file_path

    caplog.clear()


def test_configure_runtime(h5py_molecules_file, config_toml,
                           change_tmp_dir, caplog):
    caplog.set_level(logging.ERROR)
    comm = MPI.COMM_WORLD
    config_toml_file, _ = config_toml

    basearg = [config_toml_file, h5py_molecules_file]

    parsed, config, prng, topol = configure_runtime(basearg, comm)
    assert isinstance(parsed, Namespace)
    assert isinstance(config, Config)
    assert isinstance(prng, np.random.Generator)
    assert parsed.destdir == "."
    assert parsed.logfile == "sim.log"
    assert parsed.plumed_outfile == "plumed.out"

    parsed, _, _, _ = configure_runtime(["-v", "2"]+basearg, comm)
    assert parsed.verbose == 2

    parsed, _, _, _ = configure_runtime(["--disable-field"]+basearg, comm)
    assert parsed.disable_field

    parsed, _, _, _ = configure_runtime(["--disable-bonds"]+basearg, comm)
    assert parsed.disable_bonds

    parsed, _, _, _ = configure_runtime(["--disable-angle-bonds"]+basearg, comm)
    assert parsed.disable_angle_bonds

    parsed, _, _, _ = configure_runtime(["--disable-dihedrals"]+basearg, comm)
    assert parsed.disable_dihedrals

    parsed, _, _, _ = configure_runtime(["--disable-dipole"]+basearg, comm)
    assert parsed.disable_dipole

    parsed, _, _, _ = configure_runtime(["--double-precision"]+basearg, comm)
    assert parsed.double_precision

    parsed, _, _, _ = configure_runtime(["--double-output"]+basearg, comm)
    assert parsed.double_output

    parsed, _, _, _ = configure_runtime(["--dump-per-particle"]+basearg, comm)
    assert parsed.dump_per_particle

    parsed, _, _, _ = configure_runtime(["--force-output"]+basearg, comm)
    assert parsed.force_output

    parsed, _, _, _ = configure_runtime(["--velocity-output"]+basearg, comm)
    assert parsed.velocity_output

    parsed, _, _, _ = configure_runtime(["--disable-mpio"]+basearg, comm)
    assert parsed.disable_mpio

    parsed, _, _, _ = configure_runtime(["--destdir", "testdir"]+basearg, comm)
    assert parsed.destdir == "testdir"

    parsed, _, _, _ = configure_runtime(["--seed", "54321"]+basearg, comm)
    assert parsed.seed == 54321

    parsed, _, _, _ = configure_runtime(["--logfile", "test.log"]+basearg, comm)
    assert parsed.logfile == "test.log"

    parsed, _, _, _ = configure_runtime(["--plumed", "test.log"]+basearg, comm)
    assert parsed.plumed == "test.log"

    parsed, _, _, _ = configure_runtime(
                       ["--plumed-outfile", "test.plumed.out"]+basearg,
                       comm
                   )
    assert parsed.plumed_outfile == "test.plumed.out"




import pytest
import logging
import os
from mpi4py import MPI
from hymd.logger import Logger, get_version, print_header
from hymd.version import __version__

@pytest.mark.mpi()
def test_logger(tmp_path):
    logname = os.path.join(f"{tmp_path}","hymd_test.log")

    # broadcast filename to all ranks
    logname = MPI.COMM_WORLD.bcast(logname, root=0)

    # Setup logger
    Logger.setup(
        default_level=logging.INFO,
        log_file=logname,
        verbose=1
    )

    MPI.COMM_WORLD.Barrier()

    assert Logger.log_file == logname

    # test
    Logger.rank0.log(logging.INFO, "TEST INF RANK0")
    MPI.COMM_WORLD.Barrier()
    with open(logname, 'r') as f:
        logcontent = f.readlines()
        assert " INFO " in logcontent[-1]
        assert "test_logger.py" in logcontent[-1]
        assert "<test_logger>" in logcontent[-1]
        assert "rank 0" in logcontent[-1]
        assert "TEST INF RANK0"in logcontent[-1]

    Logger.all_ranks.log(logging.INFO, "TEST INF ALL_RANKS")
    MPI.COMM_WORLD.Barrier()
    with open(logname, 'r') as f:
        logcontent = f.readlines()
        assert "TEST INF ALL_RANKS" in logcontent[-1]

    Logger.rank0.log(logging.WARNING, "TEST WARN RANK0")
    MPI.COMM_WORLD.Barrier()
    with open(logname, 'r') as f:
        logcontent = f.readlines()
        assert " WARNING " in logcontent[-1]
        assert "TEST WARN RANK0" in logcontent[-1]

    Logger.rank0.log(logging.ERROR, "TEST ERR RANK0")
    MPI.COMM_WORLD.Barrier()
    with open(logname, 'r') as f:
        logcontent = f.readlines()
        assert " ERROR " in logcontent[-1]
        assert "TEST ERR RANK0" in logcontent[-1]


def test_version():
    try:
      import git
      foundgit = True
    except:
      foundgit = False

    version = get_version()

    if foundgit:
        assert 'branch commit' in version
    
    assert __version__ in version


def test_header():
    header = print_header()

    assert __version__ in header



"""Implements the Berendsen barostat.
Scales the box and particle positions during simulation to
simulate coupling to an external pressure bath set at a
target pressure.

It calculates the scaling factor according to:
.. math::

    \\alpha_{L,N} = 1 - \\frac{dt n_b}{\\tau_p}\\β(P_{L,N}^t - P_{L,N})

where :math:`dt` is the outer rRESPA time-step, :math:`n_b` is the frequency
of barostat calls, :math:`\\tau_p` is the pressure coupling time constant,
:math:`\\beta` is the isothermal compressibility, :math:`P_{L,N}^t` and 
:math:`P_{L,N}` is the target and instantaneous internal pressure in the 
lateral (L) and normal (N) directions respectively. Convention: Cartesian
z-direction is considered normal.

The box and particle positions are scaled in the L and N directions according
to the nature of the barostat (see functions `isotropic` and `semiisotropic`
below by an amount :math:`\α^{\\frac{1}{3}}`.

The updated system information is passed on to the pmesh objects.

References
----------
H. J. C. Berendsen, J. P. M. Postma, W. F. van Gunsteren,
A. DiNola, and J. R. Haak , "Molecular dynamics with coupling
to an external bath", J. Chem. Phys. 81, 3684-3690 (1984)
"""
import numpy as np
from mpi4py import MPI
from dataclasses import dataclass
from typing import Union
from .pressure import comp_pressure
from .field import initialize_pm


@dataclass
class Target_pressure:
    P_L: Union[bool, float]
    P_N: Union[bool, float]


def isotropic(
    pmesh,
    pm_stuff,
    phi,
    phi_q,
    psi,
    hamiltonian,
    positions,
    velocities,
    config,
    phi_fft,
    phi_laplacian,
    phi_transfer,
    bond_pr,
    angle_pr,
    step,
    prng,
    comm=MPI.COMM_WORLD,
):
    """
    Implements an isotropic Berendsen barostat.
    The box and particle positions are scaled uniformly
    in the L and N directions.

    Parameters
    ----------
    pmesh : module 'pmesh.pm'
    pm_stuff : list[Union(pmesh.pm.RealField, pmesh.pm.ComplexField]
        List of pmesh objects.
    phi : list[pmesh.pm.RealField], (M,)
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type
        :code:`M`. Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    velocities : (N, D) numpy.ndarray
        Array of velocities of N particles in D dimensions.
    config : Config
        Configuration dataclass containing simulation metadata and parameters.
    phi_fft : list[pmesh.pm.ComplexField], (M,)
        Pmesh :code:`ComplexField` objects containing discretized particle
        number density values in reciprocal space on the computational grid;
        one for each particle type. Pre-allocated, but empty; any values in
        this field are discarded Changed in-place. Local for each MPI rank--the
        full computaional grid is represented by the collective fields of all
        MPI ranks.
    phi_laplacian : list[pmesh.pm.RealField], (M, 3)
        Like phi, but containing the laplacian of particle number densities.
    phi_transfer : list[pmesh.pm.ComplexField], (3,)
        Like phi_fourier, used as an intermediary to perform FFT operations
        to obtain the gradient or laplacian of particle number densities.
    bond_pr : (3,) numpy.ndarray
        Total bond pressure due all two-particle bonds.
    angle_pr : (3,) numpy.ndarray
        Total angle pressure due all three-particle bonds.
    step : integer
        MD step number
    prng : np.random.Generator
        Numpy object that provides a stream of random bits
    comm : MPI.Intracomm, optional
        MPI communicator to use for rank commuication. Defaults to
        MPI.COMM_WORLD.

    Returns
    -------
    pm_stuff : list[Union(pmesh.pm.RealField, pmesh.pm.ComplexField]
        List of modified/unmodified pmesh objects.
    change : Boolean
        Indicates whether or not any pmesh objects were reinitialized.
    """
    rank = comm.Get_rank()
    beta = 4.6 * 10 ** (-5)  # bar^(-1) #isothermal compressibility of water
    change = False

    if np.mod(step, config.n_b) == 0:
        change = True
        # compute pressure
        pressure = comp_pressure(
            phi,
            phi_q,
            psi,
            hamiltonian,
            velocities,
            config,
            phi_fft,
            phi_laplacian,
            phi_transfer,
            positions,
            bond_pr,
            angle_pr,
            comm=comm,
        )

        # Total pressure across all ranks
        P = np.average(pressure[-3:-1])  # kJ/(mol nm^3)
        P = P * 16.61  # bar

        # scaling factor
        alpha = (
            1.0
            - (config.time_step * config.respa_inner)
            * config.n_b
            / config.tau_p
            * beta
            * (config.target_pressure.P_L - P)
        ) ** (1 / 3)

        # length scaling
        config.box_size *= alpha

        # position coordinates scaling
        positions *= alpha

        # pmesh re-initialize
        pm_stuff = initialize_pm(pmesh, config, comm)
    return (pm_stuff, change)


def semiisotropic(
    pmesh,
    pm_stuff,
    phi,
    phi_q,
    psi,
    hamiltonian,
    positions,
    velocities,
    config,
    phi_fft,
    phi_laplacian,
    phi_transfer,
    bond_pr,
    angle_pr,
    step,
    prng,
    comm=MPI.COMM_WORLD,
):
    """
    Implements a semiisotropic Berendsen barostat.
    The box and particle positions are scaled by :math:`\\alpha_L^{\\frac{1}{3}}`
    in the L direction and by :math:`\\alpha_N^{\\frac{1}{3}}` in the N direction.

    Parameters
    ----------
    pmesh : module 'pmesh.pm'
    pm_stuff : list[Union(pmesh.pm.RealField, pmesh.pm.ComplexField]
        List of pmesh objects.
    phi : list[pmesh.pm.RealField], (M,)
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type
        :code:`M`. Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    velocities : (N, D) numpy.ndarray
        Array of velocities of N particles in D dimensions.
    config : Config
        Configuration dataclass containing simulation metadata and parameters.
    phi_fft : list[pmesh.pm.ComplexField], (M,)
        Pmesh :code:`ComplexField` objects containing discretized particle
        number density values in reciprocal space on the computational grid;
        one for each particle type. Pre-allocated, but empty; any values in
        this field are discarded Changed in-place. Local for each MPI rank--the
        full computaional grid is represented by the collective fields of all
        MPI ranks.
    phi_laplacian : list[pmesh.pm.RealField], (M, 3)
        Like phi, but containing the laplacian of particle number densities.
    phi_transfer : list[pmesh.pm.ComplexField], (3,)
        Like phi_fourier, used as an intermediary to perform FFT operations
        to obtain the gradient or laplacian of particle number densities.
    bond_pr : (3,) numpy.ndarray
        Total bond pressure due all two-particle bonds.
    angle_pr : (3,) numpy.ndarray
        Total angle pressure due all three-particle bonds.
    step : integer
        MD step number
    prng : np.random.Generator
        Numpy object that provides a stream of random bits
    comm : MPI.Intracomm, optional
        MPI communicator to use for rank commuication. Defaults to
        MPI.COMM_WORLD.

    Returns
    -------
    pm_stuff : list[Union(pmesh.pm.RealField, pmesh.pm.ComplexField]
        List of modified/unmodified pmesh objects.
    change : Boolean
        Indicates whether or not any pmesh objects were reinitialized.
    """
    rank = comm.Get_rank()
    beta = 4.6 * 10 ** (-5)  # bar^(-1) #isothermal compressibility of water
    change = False
    if np.mod(step, config.n_b) == 0:
        change = True

        # compute pressure
        pressure = comp_pressure(
            phi,
            phi_q,
            psi,
            hamiltonian,
            velocities,
            config,
            phi_fft,
            phi_laplacian,
            phi_transfer,
            positions,
            bond_pr,
            angle_pr,
            comm=comm,
        )

        # Total pressure across all ranks
        # L: Lateral; N: Normal
        [PL, PN] = [0, 0]
        PL = (pressure[-3] + pressure[-2]) / 2  # kJ/(mol nm^3)
        PN = pressure[-1]  # kJ/(mol nm^3)
        PL = PL * 16.61  # bar
        PN = PN * 16.61  # bar
        alphaL = 1.0
        alphaN = 1.0

        if config.target_pressure.P_L:
            # scaling factor
            alphaL = (
                1.0
                - (config.time_step * config.respa_inner)
                * config.n_b
                / config.tau_p
                * beta
                * (config.target_pressure.P_L - PL)
            ) ** (1 / 3)
            # length scaling
            config.box_size[0:2] *= alphaL

            positions[:][0:2] *= alphaL

        if config.target_pressure.P_N:
            # scaling factor
            alphaN = (
                1.0
                - (config.time_step * config.respa_inner)
                * config.n_b
                / config.tau_p
                * beta
                * (config.target_pressure.P_N - PN)
            ) ** (1 / 3)
            # length scaling
            config.box_size[2] *= alphaN

            positions[:][2] *= alphaN

        # pmesh re-initialize
        pm_stuff = initialize_pm(pmesh, config, comm)
    return (pm_stuff, change)



"""Handles event logging for simulation information, warnings, and errors.
"""
import sys
import os
import logging
from mpi4py import MPI
from .version import __version__


class MPIFilterRoot(logging.Filter):
    """Log output Filter wrapper class for the root MPI rank log"""

    def filter(self, record):
        """Log event message filter

        Parameters
        ----------
        record : logging.LogRecord
            LogRecord object corresponding to the log event.
        """
        if record.funcName == "<module>":
            record.funcName = "main"
        if MPI.COMM_WORLD.Get_rank() == 0:
            record.rank = MPI.COMM_WORLD.Get_rank()
            record.size = MPI.COMM_WORLD.Get_size()
            return True
        else:
            return False


class MPIFilterAll(logging.Filter):
    """Log output Filter wrapper class for the all-MPI-ranks log"""

    def filter(self, record):
        """Log event message filter

        Parameters
        ----------
        record : logging.LogRecord
            LogRecord object corresponding to the log event.
        """
        if record.funcName == "<module>":
            record.funcName = "main"
        record.rank = MPI.COMM_WORLD.Get_rank()
        record.size = MPI.COMM_WORLD.Get_size()
        return True


class Logger:
    """Log output handler class

    Notes
    -----
    This wraps the default python library :code:`logging`, see
    `docs.python.org/3/library/logging.html`_.

    .. _`docs.python.org/3/library/logging.html`:
        https://docs.python.org/3/library/logging.html

    Attributes
    ----------
    level : int
        Determines the verbosity level of the log, corresponding to
        logging.level. Numerical values :code:`50` (:code:`logging.CRITICAL`),
        :code:`40` (:code:`logging.ERROR`), :code:`30`
        (:code:`logging.WARNING`), :code:`20` (:code:`logging.INFO`),
        :code:`10` (:code:`logging.DEBUG`), and :code:`0`
        (:code:`logging.UNSET`) are supported values. Any log event message
        less severe than the specified level is ignored. All other event
        messages are emitted.
    log_file : str
        Path to output log file.
    format : str
        Prepended dump string for each log event. Specifies the log event
        level, the module emitting the event, the code line, the enclosing
        function name, and the MPI rank writing the message.
    date_format : str
        Prepends the date before all log event messages.
    formatter : logging.Formatter
        Formatter handling the prepending of the information in `format` and
        `date_format` to each log event message. Used by default for all
        loggers.
    rank0 : logging.Logger
        Default logger object for the root MPI rank.
    all_ranks : logging.Logger
        Default logger object for messages being emitted from all MPI ranks
        simultaneously.
    """

    level = None
    log_file = None
    format = " %(levelname)-8s [%(filename)s:%(lineno)d] <%(funcName)s> {rank %(rank)d/%(size)d} %(message)s"  # noqa: E501
    date_format = "%(asctime)s"
    formatter = logging.Formatter(fmt=date_format + format)
    rank0 = logging.getLogger("HyMD.rank_0")
    all_ranks = logging.getLogger("HyMD.all_ranks")

    @classmethod
    def setup(cls, default_level=logging.INFO, log_file=None, verbose=False):
        """Sets up the logger object.

        If a :code:`log_file` path is provided, log event messages are output
        to it. Otherwise, the logging messages are emitted to stdout.

        Parameters
        ----------
        default_level : int, optional
            Default verbosity level of the logger. Unless specified, it is
            :code:`10` (:code:`logging.INFO`).
        log_file : str, optional
            Path to output log file. If `None` or not priovided, no log file is
            used and all logging is done to stdout.
        verbose : bool, optional
            Increases the logging level to :code:`30` (:code:`logging.WARNING`)
            if True, otherwise leaves the logging level unchanged.
        """
        cls.level = default_level
        cls.log_file = log_file

        log_to_stdout = False
        if verbose:
            log_to_stdout = True

        level = default_level
        if not verbose:
            level = logging.WARNING
        cls.rank0.setLevel(level)
        cls.all_ranks.setLevel(level)

        cls.rank0.addFilter(MPIFilterRoot())
        cls.all_ranks.addFilter(MPIFilterAll())

        if (not log_file) and (not log_to_stdout):
            return
        if log_file:
            cls.log_file_handler = logging.FileHandler(log_file)
            cls.log_file_handler.setLevel(level)
            cls.log_file_handler.setFormatter(cls.formatter)

            cls.rank0.addHandler(cls.log_file_handler)
            cls.all_ranks.addHandler(cls.log_file_handler)
        if log_to_stdout:
            cls.log_to_stdout = True
            cls.stdout_handler = logging.StreamHandler()
            cls.stdout_handler.setLevel(level)
            cls.stdout_handler.setStream(sys.stdout)
            cls.stdout_handler.setFormatter(cls.formatter)

            cls.rank0.addHandler(cls.stdout_handler)
            cls.all_ranks.addHandler(cls.stdout_handler)


def format_timedelta(timedelta):
    days = timedelta.days
    hours, rem = divmod(timedelta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)
    microseconds = timedelta.microseconds
    ret_str = ""
    if days != 0:
        ret_str += f"{days} days "
    ret_str += f"{hours:02d}:{minutes:02d}:{seconds:02d}.{microseconds:06d}"
    return ret_str


def get_version():
    # Check if we are in a git repo and grab the commit hash and the branch if
    # we are, append it to the version number in the output specification.
    try:
        import git

        try:
            repo_dir = os.path.abspath(
                os.path.join(os.path.dirname(__file__), os.pardir)
            )
            repo = git.Repo(repo_dir)
            commit = str(repo.head.commit)[:7]
            branch = repo.active_branch
            version = f"{__version__} [{branch} branch commit {commit}]"
        except git.exc.InvalidGitRepositoryError:
            version = f"{__version__}"
    except ModuleNotFoundError:
        version = f"{__version__}"

    return version


def print_header():
    banner = r"""
     __  __      ____                              __  _______ 
    / / / /_  __/ / /__  _________ _____ ______   /  |/  / __ \
   / /_/ / / / / / / _ \/ ___/ __ `/ __ `/ ___/  / /|_/ / / / /
  / __  / /_/ / / /  __/ /  / /_/ / /_/ (__  )  / /  / / /_/ / 
 /_/ /_/\__, /_/_/\___/_/   \__,_/\__,_/____/  /_/  /_/_____/  
       /____/ 
    """

    refs_set = """
 
 [1] Ledum, M.; Sen, S.; Li, X.; Carrer, M.; Feng Y.; Cascella, M.; Bore, S. L. 
 HylleraasMD: A Domain Decomposition-Based Hybrid Particle-Field Software for Multi-Scale Simulations of Soft Matter.
 J. Chem. Theory Comput. 2023.

 [2] Ledum, M.; Carrer, M.; Sen, S.; Li, X.; Cascella, M.; Bore, S. L. 
 HyMD: Massively parallel hybrid particle-field molecular dynamics in Python.
 J. Open Source Softw. 8(84), 4149, 2023.
 
 [3] Sen, S.; Ledum, M.; Bore, S. L.; Cascella, M. 
 Soft Matter under Pressure: Pushing Particle–Field Molecular Dynamics to the Isobaric Ensemble.
 J Chem Inf Model 2023, 63(7), 1549-9596.

 [4] Bore, S. L.; Cascella, M. 
 Hamiltonian and alias-free hybrid particle–field molecular dynamics.
 J. Chem. Phys. 2020, 153, 094106.

 [5] Pippig, M. PFFT: An extension of FFTW to massively parallel architectures.
 SIAM J. Sci. Comput. 2013, 35, C213–C236.

"""

    version = f"Version {get_version()}"
    header = banner
    header += version.center(56) + "\n\n"
    #header += " Please read and cite the references below:"
    header += " PLEASE READ AND CITE THE REFERENCES BELOW:"
    header += refs_set

    return header



from mpi4py import MPI
import pytest
import logging
import pmesh
import numpy as np
from hymd.hamiltonian import (
    Hamiltonian, DefaultNoChi, DefaultWithChi, SquaredPhi, get_hamiltonian,
)
from hymd.input_parser import (
    _find_unique_names, Config, read_config_toml, parse_config_toml,
)
from hymd.force import Chi


@pytest.mark.parametrize(
    "sigma",
    [
        0.2988365823859701,
        1.2585762493242553,
        9.2159828579248931,
    ],
)
def test_DefaultNoChi_window_function(
    dppc_single, config_toml, filter, caplog
):
    caplog.set_level(logging.INFO)
    sigma, filter = filter
    # print(sigma, filter)
    indices, _, names, _, r, _ = dppc_single
    conf_file_name, _ = config_toml
    file_contents = read_config_toml(conf_file_name)
    config = parse_config_toml(file_contents)

    config = _find_unique_names(config, names)
    config.sigma = sigma
    config.type_charges = [0.] * config.n_types

    W = DefaultNoChi(config)

    k_ = np.array(
        [[0.5321315378106508, -0.6711591309063634,  0.8362051282174443],
         [0.2853046917286570, -0.6542962742862817,  0.3174805390299977],
         [0.6999748102259762, -0.9385345654631219, -0.7383831543700541]]
    )
    v = np.array(
        [[0.6106860760556785, -0.5406324770662296, 0.6388756736156205],
         [-0.7348831910188103, 0.2808258965802970, 0.7446817693106476],
         [0.6458163432308923, -0.0526126093343278, 0.7065510160449484]]
    )

    for kk, vv in zip(k_, v):
        assert np.allclose(W.H(kk, vv), filter(kk, vv), atol=1e-14)


@pytest.mark.parametrize(
    ["types", "kappa", "rho0", "sigma"],
    [
        (["A", "B"], 0.029230985982, 0.012169633583017335, 0.2988365823859701),
        (["A", "B"], 1.299759825895, 0.012169633583017335, 1.2095870248085025),
    ],
)
def test_DefaultNoChi_energy_functional(filter, v_ext, caplog):
    _, filter = filter
    types, n_types, kappa, rho0, sigma, w, v_external = v_ext

    caplog.set_level(logging.INFO)
    config_conf = Config(
        n_particles=5,
        mesh_size=[8, 8, 8],
        box_size=[7.1598, 11.2498, 5.1009],
        time_step=0.01,
        n_steps=1,
        sigma=sigma,
        kappa=kappa,
    )
    r = np.array(
        [[5.0856957005460090, 2.6917109262984917, 2.1680859811719504],
         [0.0592355479043435, 7.4705053874595030, 0.7169198724514365],
         [6.8662344320308580, 3.1065309079252390, 3.5621049784729850],
         [5.5405900785250210, 6.5585020954226065, 1.3155113935992528],
         [1.6400925628903045, 2.1325438112867725, 3.4149968975443680]]
    )
    types = np.array([0, 0, 1, 1, 1], dtype=int)
    names = np.array([np.string_(x) for x in ["A", "A", "B", "B", "B"]])
    n_types = len(np.unique(types))
    config_conf = _find_unique_names(config_conf, names)
    config_conf.type_charges = [0.] * config_conf.n_types

    W = DefaultNoChi(config_conf)

    mesh_size = config_conf.mesh_size
    pm = pmesh.ParticleMesh(
        mesh_size, BoxSize=config_conf.box_size, dtype='f8',
        comm=MPI.COMM_WORLD
    )
    layouts = [pm.decompose(r[t == types]) for t in
               range(len(config_conf.unique_names))]
    painted = [
        pm.paint(r[t == types], layout=layouts[t])
        for t in range(len(config_conf.unique_names))
    ]
    painted_ = [
        pm.paint(r[t == types], layout=layouts[t]) for t in range(n_types)
    ]
    # Particles -> grid operation should conserve total mass
    assert np.sum(np.sum(painted[0])) == pytest.approx(2.0, abs=1e-14)
    assert np.sum(np.sum(painted_[0])) == pytest.approx(2.0, abs=1e-14)
    assert np.sum(np.sum(painted[1])) == pytest.approx(3.0, abs=1e-14)
    assert np.sum(np.sum(painted_[1])) == pytest.approx(3.0, abs=1e-14)

    pm = pmesh.ParticleMesh(
        config_conf.mesh_size, BoxSize=np.array(config_conf.box_size),
        dtype="f8", comm=MPI.COMM_WORLD
    )

    for p in painted:
        p.r2c(out=Ellipsis).apply(W.H, out=Ellipsis).c2r(out=Ellipsis)
    for p_ in painted_:
        p_.r2c(out=Ellipsis).apply(filter, out=Ellipsis).c2r(out=Ellipsis)
    for p, p_ in zip(painted, painted_):
        assert np.allclose(p, p_, atol=1e-14)

    # Particles -> grid -> FFT -> convolution with filter H -> iFFT operation
    # should conserve total mass
    assert np.sum(np.sum(painted[0])) == pytest.approx(2.0, abs=1e-14)
    assert np.sum(np.sum(painted_[0])) == pytest.approx(2.0, abs=1e-14)
    assert np.sum(np.sum(painted[1])) == pytest.approx(3.0, abs=1e-14)
    assert np.sum(np.sum(painted_[1])) == pytest.approx(3.0, abs=1e-14)

    filtered = painted
    filtered_ = painted_

    for t in range(len(config_conf.unique_names)):
        v_ext_fourier_space = (
            W.v_ext[t](filtered).r2c(out=Ellipsis).apply(W.H, out=Ellipsis)
        )
        v_ext_fourier_space_ = (
            v_external[t](filtered_).r2c(out=Ellipsis)
                                    .apply(filter, out=Ellipsis)
        )
        assert np.allclose(v_ext_fourier_space, v_ext_fourier_space_, atol=1e-14)  # noqa: E501

    V = np.prod(config_conf.box_size)
    n_mesh__cells = np.prod(np.full(3, config_conf.mesh_size))
    volume_per_cell = V / n_mesh__cells
    painted = [
        pm.paint(r[t == types], layout=layouts[t]) / volume_per_cell
        for t in range(len(config_conf.unique_names))
    ]
    painted_ = [
        pm.paint(r[t == types], layout=layouts[t]) / volume_per_cell
        for t in range(n_types)
    ]
    for p in painted:
        p.r2c(out=Ellipsis).apply(W.H, out=Ellipsis).c2r(out=Ellipsis)
    for p_ in painted_:
        p_.r2c(out=Ellipsis).apply(filter, out=Ellipsis).c2r(out=Ellipsis)
    w__ = W.w(painted) * volume_per_cell
    w_ = w(painted_) * volume_per_cell

    assert np.allclose(w__, w_, atol=1e-14)
    assert w__.csum() == pytest.approx(w_.csum(), abs=1e-14)


@pytest.mark.parametrize(
    ["types", "kappa", "rho0", "sigma"],
    [
        (["A"], 0.029230985982, 0.000888888888888888, 0.2988365823859701),
        (["A"], 1.299759825895, 0.000888888888888888, 1.2095870248085025),
    ],
)
def test_Hamiltonian_no_chi_gaussian_core(v_ext, caplog):
    caplog.set_level(logging.INFO)
    types, n_types, kappa, rho0, sigma, w, v_external = v_ext

    config = Config(
        n_particles=3,
        mesh_size=[160, 160, 160],
        box_size=[15.0, 15.0, 15.0],
        time_step=0.01,
        n_steps=1,
        sigma=sigma,
        kappa=kappa,
    )
    names = np.array([np.string_(s) for s in ['A', 'A']])
    config = _find_unique_names(config, names)
    config.type_charges = [0.] * config.n_types
    r = np.array(
        [[1.50, 0.75, 2.25],
         [2.25, 0.00, 3.00],
         [4.50, 1.50, 2.25]]
    )
    pm = pmesh.ParticleMesh(
        config.mesh_size, BoxSize=config.box_size, dtype="f8",
        comm=MPI.COMM_WORLD
    )
    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells

    layout = pm.decompose(r)
    phi = pm.paint(r, layout=layout) / volume_per_cell
    assert phi.csum() == pytest.approx(config.n_particles / volume_per_cell, abs=1e-14)  # noqa: E501
    hamiltonian = Hamiltonian(config)
    phi = (phi.r2c(out=Ellipsis)
              .apply(hamiltonian.H, out=Ellipsis)
              .c2r(out=Ellipsis))
    assert phi.csum() == pytest.approx(config.n_particles / volume_per_cell, abs=1e-11)  # noqa: E501

    for W in [SquaredPhi(config), DefaultNoChi(config)]:
        w = (W.w([phi]) * volume_per_cell).csum()

        # Gaussian core model energy
        pi32 = np.arccos(-1.0)**(3.0 / 2.0)
        c = 16.0 * pi32 * config.kappa * config.sigma**3 * config.rho0
        diag = config.n_particles / c
        offdiag = 0
        for i in range(config.n_particles):
            for j in range(i + 1, config.n_particles):
                rij = r[i, :] - r[j, :]
                rij2 = np.dot(rij, rij)
                offdiag += 2 * np.exp(- rij2 / (4.0 * config.sigma**2)) / c
        if isinstance(W, DefaultNoChi):
            E = diag + offdiag - 0.5 * config.n_particles / config.kappa
        else:
            E = diag + offdiag
        assert w == pytest.approx(E, abs=1e-6)

    # three particles, W=phi^2/(2 phi0 kappa)
    # 27046.447859682066 + 160.09645705998503 = 27206.544316742053 -->const+exp
    # 27206.544316742053 --> mathematica
    # 27206.544316741987 --> grid 640
    # 27206.544316741983 --> grid 320
    # 27206.544316741983 --> grid 160

    # two particles, W=phi^2/(2 phi0 kappa)
    # 27046.44785968207 + 240.14467831817544 = 27286.592538000245 -->const+exp
    # 27286.592538000245 --> mathematica
    # 27286.592538000346 --> grid 640
    # 27286.592538000346 --> grid 320
    # 27286.59253800035 --> grid 160

    # two particles, W=phi^2
    # 1.6823405640794202 + 0.014937456322488748 = 1.6972780204 -->const+exp
    # 1.697278020401909 --> mathematica
    # 1.6972780204019127 --> grid 640
    # 1.6972780204019127 --> grid 320
    # 1.6972780204019127 --> grid 160

    # one particle, W=phi^2/(2 phi0 kappa)
    # 27046.44785968207 --> mathematica
    # 27046.447859682085 --> grid 640
    # 27046.447859682085 --> grid 320
    # 27046.447859682088 --> grid 160

    # one particle, W=phi^2
    # 0.8411702820397101 --> mathematica
    # 0.841170282039712 --> grid 640
    # 0.841170282039712  --> grid 320
    # 0.8411702820397121 --> grid 160


@pytest.mark.parametrize(
    ["types", "kappa", "rho0", "sigma"],
    [
        (["A", "B", "C"], 0.029230985982, 0.0014814814814814814, 0.2988365823859701),  # noqa: E501
        (["A", "B", "C"], 1.299759825895, 0.0014814814814814814, 1.2095870248085025),  # noqa: E501
    ],
)
def test_Hamiltonian_with_chi_gaussian_core(v_ext, caplog):
    caplog.set_level(logging.INFO)
    types, n_types, kappa, rho0, sigma, w, v_external = v_ext
    config = Config(
        n_particles=5,
        mesh_size=[160, 160, 160],
        box_size=[15.0, 15.0, 15.0],
        time_step=0.01,
        n_steps=1,
        sigma=sigma,
        kappa=kappa,
    )

    names = np.array([np.string_(s) for s in ['A', 'A', 'B', 'C', 'C']])
    types = np.array([0, 0, 1, 2, 2])
    config = _find_unique_names(config, names)
    config.box_size = np.array([15.0, 15.0, 15.0])
    config.mesh_size = np.array([160, 160, 160])
    config.n_particles = 5
    config.type_charges = [0.] * config.n_types

    r = np.array(
        [[1.50, 0.75, 2.25],
         [2.25, 0.00, 3.00],
         [4.50, 1.50, 2.25],
         [1.50, 1.50, 0.75],
         [3.00, 4.50, 1.50]]
    )
    pm = pmesh.ParticleMesh(
        config.mesh_size, BoxSize=config.box_size, dtype="f8",
        comm=MPI.COMM_WORLD
    )
    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells

    layouts = [pm.decompose(r[types == t]) for t in range(config.n_types)]
    phi = [pm.paint(r[types == t], layout=layouts[t]) / volume_per_cell
           for t in range(config.n_types)]
    hamiltonian = Hamiltonian(config)
    for t in range(config.n_types):
        phi[t].r2c(out=Ellipsis).apply(hamiltonian.H, out=Ellipsis).c2r(out=Ellipsis)  # noqa: E501

    chi_ = [
      [['A', 'B'],   [9.6754032616815161]],
      [['A', 'C'], [-13.2596290315913623]],
      [['B', 'C'],   [0.3852001771213374]]
    ]
    chi_dict = {("A", "B"):   9.6754032616815161,
                ("A", "C"): -13.2596290315913623,
                ("B", "C"):   0.3852001771213374}
    chi = [None] * ((config.n_types - 1) * config.n_types // 2)
    for i, c in enumerate(chi_):
        chi[i] = Chi(
            atom_1=c[0][0], atom_2=c[0][1], interaction_energy=c[1][0]
        )
    config.chi = chi
    type_to_name_map = {0: 'A', 1: 'B', 2: 'C'}

    W = DefaultWithChi(config, config.unique_names, type_to_name_map)
    W_ = DefaultNoChi(config)
    w = (W.w(phi) * volume_per_cell).csum()
    w_ = (W_.w(phi) * volume_per_cell).csum()

    # Gaussian core model energy
    pi32 = np.arccos(-1.0)**(3.0 / 2.0)
    c = 16.0 * pi32 * config.kappa * config.sigma**3 * config.rho0
    diag = config.n_particles / c
    offdiag = 0
    for i in range(config.n_particles):
        for j in range(i + 1, config.n_particles):
            rij = r[i, :] - r[j, :]
            rij2 = np.dot(rij, rij)
            offdiag += 2 * np.exp(- rij2 / (4.0 * config.sigma**2)) / c
    E = diag + offdiag - 0.5 * config.n_particles / config.kappa
    assert w_ == pytest.approx(E, abs=1e-6)

    interaction_energy = 0
    for i in range(config.n_particles):
        for j in range(i + 1, config.n_particles):
            ni = names[i].decode('utf-8')
            nj = names[j].decode('utf-8')
            if ni != nj:
                rij = r[i, :] - r[j, :]
                rij2 = np.dot(rij, rij)
                c_ = 2 * config.kappa * chi_dict[tuple(sorted([ni, nj]))] / c
                interaction_energy += (
                    c_ * np.exp(-rij2 / (4.0 * config.sigma**2))
                )
    E = E + interaction_energy
    assert w == pytest.approx(E, abs=1e-6)

    # five particles, three types, W=sum chi(i,j) phi(i) phi(j)
    # -2.8663135769749566  --> mathematica
    # -2.8663135769747057 --> grid 640
    # -2.866313576974921 --> grid 320
    # -2.866313576974796 --> grid 160


def test_get_hamiltonian(config_toml):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)
    config.unique_names = ["N","P","G","C","W"]
    config.type_to_name_map = {0: 'N', 1: 'P', 2: 'G', 3: 'C', 4: 'W'}
    config.n_types = 5
    config.type_charges = [0., 0., 0., 0., 0.]


    config.hamiltonian = "DefaultNoChi"
    hamiltonian = get_hamiltonian(config)
    assert isinstance(hamiltonian, DefaultNoChi)

    config.hamiltonian = "DefaultWithChi"
    hamiltonian = get_hamiltonian(config)
    assert isinstance(hamiltonian, DefaultWithChi)

    config.hamiltonian = "SquaredPhi"
    hamiltonian = get_hamiltonian(config)
    assert isinstance(hamiltonian, SquaredPhi)

